﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Text.RegularExpressions;
using Plux;
#if SILVERLIGHT
using System.Windows.Media;
#else
using System.Drawing;
#endif

public delegate Color ExtensionColorHander(ExtensionInfo extension);

public class DotGenerator {
  private const int DEFAULT_PLUG_POS = 2;
  private const string slotAndPlug = "<TD WIDTH=\"10\"></TD>";
  private const string plugOnly = "<TD COLSPAN=\"2\" HEIGHT=\"10\"></TD>";
  private const string slotStr = "<TD ALIGN=\"RIGHT\" BGCOLOR=\"";

  private static readonly Color PlugPointerColor = Color.FromArgb(255, 0, 0, 0); // Color.Black
  private static readonly Color ExtensionColor = Color.FromArgb(255, 220, 220, 220); // Color.Gainsboro
  private static readonly Color MarkedExentsionColor = Color.FromArgb(255, 176, 196, 222); // Color.LightSteelBlue
  private static readonly Color SlotColor = Color.FromArgb(255, 245, 245, 245); // Color.WhiteSmoke
  private static readonly Color PluxSlotColor = Color.FromArgb(255, 173, 255, 47); // Color.GreenYellow

  private readonly string extensionStr;
  private readonly string objectInfoStr;
  private readonly string blankLineStr;
  private readonly string slotOnly;
  private readonly string plugStr;

  private readonly ICollection<ExtensionInfo> extensions;
  private readonly ICollection<string> ignoredExtensions;
  private readonly ICollection<string> ignoredExtensionsRegExps;
  private readonly ICollection<int> markedExtensions;
  private readonly int plugPosition;
  private readonly ExtensionColorHander colorHandler;

  private StringBuilder sb;

  private DotGenerator(ICollection<ExtensionInfo> exts, ICollection<string> ignored, ICollection<int> marked, int plugPos, ExtensionColorHander handler) {
    if (plugPos < 0 || plugPos > 2) {
      throw new ArgumentException("PlugPos must be between 0 and 2, but was: " + plugPos);
    }

    extensions = exts;
    ignoredExtensions = ignored;
    markedExtensions = marked;
    plugPosition = plugPos;
    colorHandler = handler ?? GetExtensionColor;

    switch (plugPos) {
      case 0:
        extensionStr = "<TR><TD COLSPAN=\"2\" ALIGN=\"LEFT\">";
        objectInfoStr = "<TR><TD COLSPAN=\"2\" ALIGN=\"LEFT\">";
        blankLineStr = "<TR><TD COLSPAN=\"2\" HEIGHT=\"10\"></TD></TR>";
        // only slots in this mode
        plugStr = null;
        slotOnly = "<TD WIDTH=\"10\"></TD>";
        break;
      case 1:
        extensionStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"2\" ALIGN=\"LEFT\">";
        objectInfoStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"2\" ALIGN=\"LEFT\">";
        blankLineStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"2\" HEIGHT=\"10\"></TD></TR>";
        plugStr = "<TD ALIGN=\"LEFT\" PORT=\"P";
        slotOnly = "<TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD WIDTH=\"10\"></TD>";
        break;
      case 2:
        extensionStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"3\" ALIGN=\"LEFT\">";
        objectInfoStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"3\" ALIGN=\"LEFT\">";
        blankLineStr = "<TR><TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"3\" HEIGHT=\"10\"></TD></TR>";
        plugStr = "<TD COLSPAN=\"2\" ALIGN=\"LEFT\" PORT=\"P";
        slotOnly = "<TD HEIGHT=\"10\" BGCOLOR=\"white\"></TD><TD COLSPAN=\"2\" WIDTH=\"10\"></TD>";
        break;
    }

    //Collect regular expressions from the IgnoreList, e.g.: /^Work(*.)$/
    ignoredExtensionsRegExps = new List<string>();
    foreach (string ignoreElement in ignoredExtensions) {
      if (ignoreElement.StartsWith("/") && ignoreElement.EndsWith("/")) {
        ignoredExtensionsRegExps.Add(ignoreElement);
      }
    }
  }

  public static string GetDotString() {
    return GetDotString(null);
  }

  public static string GetDotString(ExtensionColorHander colorHander) {
    ReadOnlyCollection<ExtensionInfo> currentExtensions = Runtime.Repository.ExtensionInfos;
    ICollection<string> ignoreList = new Collection<string>();
    ICollection<int> marked = new Collection<int>();
    return GetDotString(currentExtensions, ignoreList, marked, DEFAULT_PLUG_POS, colorHander);
  }

  public static string GetDotString(ICollection<ExtensionInfo> extensions, ICollection<string> ignored, ICollection<int> marked, int plugPos) {
    return GetDotString(extensions, ignored, marked, plugPos, null);
  }

  public static string GetDotString(ICollection<ExtensionInfo> extensions, ICollection<string> ignored, ICollection<int> marked, int plugPos, ExtensionColorHander colorHander) {
    DotGenerator generator = new DotGenerator(extensions, ignored, marked, plugPos, colorHander);
    return generator.MakeDotString();
  }

  private string MakeDotString() {
    sb = new StringBuilder();

    AddHeader();
    AddExtensions();
    AddFooter();

    return sb.ToString();
  }

  private void AddHeader() {
    sb.AppendLine("digraph HotViz {");
    sb.AppendLine("rankdir=RL;");

    sb.AppendLine();

    sb.AppendLine("node [");
    sb.AppendLine("fontsize=10,");
    sb.AppendLine("fontname=\"Tahoma\",");
    sb.AppendLine("label=\"\",");
    sb.AppendLine("fontcolor=\"black\"");
    sb.AppendLine("]");

    sb.AppendLine();
  }

  private void AddFooter() {
    sb.AppendLine("}");
  }

  private void AddExtensions() {
    foreach (ExtensionInfo extension in extensions) {
      if (!isIgnoredExtension(extension)) {
        AddExtension(extension);
      }
    }
  }

  private bool isIgnoredExtension(ExtensionInfo extension) {
    if (ignoredExtensions.Contains(extension.Name)) {
      return true;
    }

    // Match the Extension's extension against Regular expressions 
    foreach (string regEx in ignoredExtensionsRegExps) {
      if (Regex.Match(extension.Name, regEx.Replace("/", String.Empty)).Success) {
        return true;
      }
    }

    return false;
  }

  private void AddExtension(ExtensionInfo extension) {
    AddExtensionHeader(extension);
    AddExtensionLabel(extension);
    AddExtensionFooter();

    AddPlugPointers(extension);
  }

  private void AddExtensionHeader(ExtensionInfo extension) {
    sb.AppendLine(GetExtensionNodeId(extension));
    sb.AppendLine(" [");
  }

  private void AddExtensionFooter() {
    sb.AppendLine("shape=plaintext");
    sb.AppendLine("]");
  }

  private void AddExtensionLabel(ExtensionInfo extension) {
    AddExtensionLabelHeader(extension);

    AddExtensionNodeLabelInfo(extension);
    AddExtensionNodeLabelSlotsAndPlugs(extension);

    AddExtensionNodeLabelFooter();
  }

  private void AddExtensionLabelHeader(ExtensionInfo extension) {
    sb.AppendLine("label=<");

    sb.Append("<TABLE PORT=\"Main\" BORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"4\" BGCOLOR=\"");
    sb.Append(ColorToHtmlString(colorHandler(extension)));
    sb.AppendLine("\">");
  }

  private void AddExtensionNodeLabelFooter() {
    sb.AppendLine("</TABLE>>,");
  }

  private void AddExtensionNodeLabelInfo(ExtensionInfo extension) {
    sb.Append(extensionStr);
    sb.Append(extension.Name);
    sb.Append(" (");
    sb.Append(extension.Id);
    sb.AppendLine(")</TD></TR>");

    if (extension.Object is IObjectInfo) {
      sb.Append(objectInfoStr);
      IObjectInfo objectInfo = (IObjectInfo)extension.Object;
      sb.Append(HTMLCharacterConverter.CreateHtmlSaveStringFrom(objectInfo.Name));
      sb.AppendLine("</TD></TR>");
    }

    sb.AppendLine(blankLineStr);
  }

  private void AddExtensionNodeLabelSlotsAndPlugs(ExtensionInfo ei) {
    if (plugPosition == 0) {
      AddOnlySlots(ei);
    } else {
      AddSlotsAndPlugs(ei);
    }
  }

  private void AddOnlySlots(ExtensionInfo extension) {
    foreach (SlotInfo slot in extension.SlotInfos) {
      sb.Append("<TR>");
      sb.Append(slotOnly);
      sb.Append(slotStr);
      sb.Append(ColorToHtmlString(GetSlotColor(slot.Name)));
      sb.Append("\" PORT=\"S");
      sb.Append(slot.Id);
      sb.Append("\">");
      if (!slot.IsOpen) {
        sb.Append("<FONT COLOR=\"RED\">");
      }
      sb.Append(slot.Name.Replace("Plux.", ""));
      sb.Append(" (");
      sb.Append(slot.Id);
      sb.Append(")");
      if (!slot.IsOpen) {
        sb.Append("</FONT>");
      }
      sb.AppendLine("</TD></TR>");
    }
    sb.AppendLine("<TR><TD COLSPAN=\"2\" HEIGHT=\"10\"></TD></TR>");
  }

  private void AddSlotsAndPlugs(ExtensionInfo ei) {
    AddSymmetricSlotsAndPlugs(ei);
    AddRestOfSlots(ei);
    AddRestOfPlugs(ei);
    sb.AppendLine(blankLineStr);
  }

  private void AddSymmetricSlotsAndPlugs(ExtensionInfo ei) {
    int nrSlotAndPlugSym = Math.Min(ei.SlotInfos.Count, ei.PlugInfos.Count);
    for (int i = 0; i < nrSlotAndPlugSym; i++) {
      sb.AppendLine("<TR>");
      AddPlug(ei.PlugInfos[i]);
      sb.Append(slotAndPlug);
      AddSlot(ei.SlotInfos[i]);
      sb.AppendLine("</TR>");
    }
  }

  private void AddRestOfSlots(ExtensionInfo ei) {
    for (int i = ei.PlugInfos.Count; i < ei.SlotInfos.Count; i++) {
      sb.AppendLine("<TR>");
      sb.Append(slotOnly);
      AddSlot(ei.SlotInfos[i]);
      sb.AppendLine("</TR>");
    }
  }

  private void AddRestOfPlugs(ExtensionInfo ei) {
    for (int i = ei.SlotInfos.Count; i < ei.PlugInfos.Count; i++) {
      sb.AppendLine("<TR>");
      AddPlug(ei.PlugInfos[i]);
      sb.Append(plugOnly);
      sb.AppendLine("</TR>");
    }
  }

  private void AddPlug(PlugInfo plug) {
    sb.Append(plugStr);
    sb.Append(plug.Id);
    sb.Append("\">");
    sb.Append(plug.Name.Replace("Plux.", ""));
    sb.Append(" (");
    sb.Append(plug.Id);
    sb.Append(")</TD>");
  }

  private void AddSlot(SlotInfo slot) {
    sb.Append(slotStr);
    sb.Append(ColorToHtmlString(GetSlotColor(slot.Name)));
    sb.Append("\" PORT=\"S");
    sb.Append(slot.Id);
    sb.Append("\">");
    if (!slot.IsOpen) {
      sb.Append("<FONT COLOR=\"RED\">");
    }
    sb.Append(slot.Name.Replace("Plux.", ""));
    sb.Append(" (");
    sb.Append(slot.Id);
    sb.Append(")");
    if (!slot.IsOpen) {
      sb.Append("</FONT>");
    }
    sb.Append("</TD>");
  }

  private void AddPlugPointers(ExtensionInfo extension) {
    string curNodeId = GetExtensionNodeId(extension);
    foreach (PlugInfo plug in extension.PlugInfos) {
      foreach (SlotInfo slot in plug.PluggedInSlots) {
        AddPlugPointer(curNodeId, plug, slot);
      }
    }
  }

  private void AddPlugPointer(string fromNodeId, PlugInfo fromPlug, SlotInfo toSlot) {
    string connection;

    if (plugPosition == 0) {
      const string format = "{0}:Main -> I{1}:S{2}:e [";
      connection = String.Format(format, fromNodeId, toSlot.ExtensionInfo.Id, toSlot.Id);
    } else {
      const string format = "{0}:P{1}:w -> I{2}:S{3}:e [";
      connection = String.Format(format, fromNodeId, fromPlug.Id, toSlot.ExtensionInfo.Id, toSlot.Id);
    }
    sb.AppendLine(connection);

    string tail = String.Format(
        "color = \"{0}\", dir = forward, arrowhead = normal, arrowtail = {1} ]",
        ColorToHtmlString(PlugPointerColor),
        toSlot.SelectedPlugs.Contains(fromPlug) ? "box" : "none");

    sb.AppendLine(tail);
  }

  private string GetExtensionNodeId(ExtensionInfo ei) {
    return "I" + ei.Id;
  }

  private static string ColorToHtmlString(Color color) {
    return String.Format("#{0:x2}{1:x2}{2:x2}", color.R, color.G, color.B);
  }

  private Color GetExtensionColor(ExtensionInfo extension) {
    return markedExtensions.Contains(extension.Id) ? MarkedExentsionColor : ExtensionColor;
  }

  private static Color GetSlotColor(string name) {
    return name.StartsWith("Plux.") ? PluxSlotColor : SlotColor;
  }
}